
/* Chrysalide - Outil d'analyse de fichiers binaires
 * hex.c - recherche de morceaux de binaire
 *
 * Copyright (C) 2023 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "hex.h"


#include <malloc.h>
#include <string.h>


#include "hex-int.h"



/* ----------------------- RECHERCHE D'UN MOTIF DE TEXTE BRUT ----------------------- */


/* Initialise la classe des recherches de texte brut. */
static void g_scan_hex_bytes_class_init(GScanHexBytesClass *klass);

/* Initialise une instance de recherche de texte brut. */
static void g_scan_hex_bytes_init(GScanHexBytes *);

/* Supprime toutes les références externes. */
static void g_scan_hex_bytes_dispose(GScanHexBytes *);

/* Procède à la libération totale de la mémoire. */
static void g_scan_hex_bytes_finalize(GScanHexBytes *);



/* --------------------- IMPLEMENTATION DES FONCTIONS DE CLASSE --------------------- */


/* Affiche un motif de recherche au format texte. */
static void g_scan_hex_bytes_output_to_text(const GScanHexBytes *, GScanContext *, int);

/* Affiche un motif de recherche au format JSON. */
static void g_scan_hex_bytes_output_to_json(const GScanHexBytes *, GScanContext *, const sized_string_t *, unsigned int, int);

/* Inscrit la définition d'un motif dans un moteur de recherche. */
//static bool g_scan_hex_bytes_enroll(GScanHexBytes *, GScanContext *, GEngineBackend *, size_t);

/* Transforme les correspondances locales en trouvailles. */
//static void g_scan_hex_bytes_check(const GScanHexBytes *, GScanContext *, GBinContent *, pending_matches_t *);



/* ---------------------------------------------------------------------------------- */
/*                         RECHERCHE D'UN MOTIF DE TEXTE BRUT                         */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour une suite d'octets à retrouver dans un binaire. */
G_DEFINE_TYPE(GScanHexBytes, g_scan_hex_bytes, G_TYPE_STRING_TOKEN);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des recherches de texte brut.           *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_hex_bytes_class_init(GScanHexBytesClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GSearchPatternClass *pattern;           /* Version de classe ancêtre   */
    GStringTokenClass *token;               /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_scan_hex_bytes_dispose;
    object->finalize = (GObjectFinalizeFunc)g_scan_hex_bytes_finalize;

    pattern = G_SEARCH_PATTERN_CLASS(klass);

    pattern->to_text = (output_pattern_to_text_fc)g_scan_hex_bytes_output_to_text;
    pattern->to_json = (output_pattern_to_json_fc)g_scan_hex_bytes_output_to_json;

    token = G_STRING_TOKEN_CLASS(klass);

    //token->enroll = (enroll_token_fc)g_scan_hex_bytes_enroll;
    //token->check = (check_token_fc)g_scan_hex_bytes_check;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : bytes = instance à initialiser.                              *
*                                                                             *
*  Description : Initialise une instance de recherche de texte brut.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_hex_bytes_init(GScanHexBytes *bytes)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : bytes = instance d'objet GLib à traiter.                     *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_hex_bytes_dispose(GScanHexBytes *bytes)
{
    G_OBJECT_CLASS(g_scan_hex_bytes_parent_class)->dispose(G_OBJECT(bytes));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : bytes = instance d'objet GLib à traiter.                     *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_hex_bytes_finalize(GScanHexBytes *bytes)
{
    G_OBJECT_CLASS(g_scan_hex_bytes_parent_class)->finalize(G_OBJECT(bytes));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : root  = représentation du motif à recherche.                 *
*                                                                             *
*  Description : Construit un gestionnaire de recherche de texte brut.        *
*                                                                             *
*  Retour      : Mécanismes mis en place.                                     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GSearchPattern *g_scan_hex_bytes_new(GScanTokenNode *root)
{
    GSearchPattern *result;                 /* Structure à retourner       */

    result = g_object_new(G_TYPE_SCAN_HEX_BYTES, NULL);

    if (!g_scan_hex_bytes_create(G_SCAN_HEX_BYTES(result), root))
        g_clear_object(&result);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : bytes = encadrement de motif à initialiser pleinement.       *
*                root  = représentation du motif à recherche.                 *
*                                                                             *
*  Description : Met en place un gestionnaire de recherche de binaire.        *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_scan_hex_bytes_create(GScanHexBytes *bytes, GScanTokenNode *root)
{
    bool result;                            /* Bilan à retourner           */

    result = g_string_token_create(G_STRING_TOKEN(bytes), root);

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                       IMPLEMENTATION DES FONCTIONS DE CLASSE                       */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : pattern = définition de motif à considérer.                  *
*                context = contexte de l'analyse à mener.                     *
*                fd      = canal d'écriture.                                  *
*                                                                             *
*  Description : Affiche un motif de recherche au format texte.               *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_hex_bytes_output_to_text(const GScanHexBytes *pattern, GScanContext *context, int fd)
{
    G_SEARCH_PATTERN_CLASS(g_scan_hex_bytes_parent_class)->to_text(G_SEARCH_PATTERN(pattern), context, fd);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : pattern = définition de motif à considérer.                  *
*                context  = contexte de l'analyse à mener.                    *
*                padding  = éventuel bourrage initial à placer ou NULL.       *
*                level    = profondeur actuelle.                              *
*                fd       = canal d'écriture.                                 *
*                                                                             *
*  Description : Affiche un motif de recherche au format JSON.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_hex_bytes_output_to_json(const GScanHexBytes *pattern, GScanContext *context, const sized_string_t *padding, unsigned int level, int fd)
{
    G_SEARCH_PATTERN_CLASS(g_scan_hex_bytes_parent_class)->to_json(G_SEARCH_PATTERN(pattern), context, padding, level, fd);

    /* TODO */

}

#if 0

/******************************************************************************
*                                                                             *
*  Paramètres  : bytes   = définition de la bribe à enregistrer.              *
*                context = contexte de l'analyse à mener.                     *
*                backend = moteur de recherche à préchauffer.                 *
*                maxsize = taille max. des atomes (mise en commun optimisée). *
*                                                                             *
*  Description : Inscrit la définition d'un motif dans un moteur de recherche.*
*                                                                             *
*  Retour      : Bilan de l'opération à renvoyer.                             *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_scan_hex_bytes_enroll(GScanHexBytes *bytes, GScanContext *context, GEngineBackend *backend, size_t maxsize)
{
    bool result;                            /* Statut à retourner          */
    size_t i;                               /* Boucle de parcours #1       */
    tracked_scan_atom_t atom;               /* Atome identifié             */
    size_t letters;                         /* Nombre de lettres présentes */
    size_t k;                               /* Boucle de parcours #2       */
    size_t extra_count;                     /* Quantité pour l'exhaustivité*/
    sized_binary_t *extra;                  /* Couverture supplémntaire    */
    size_t remaining;                       /* Quantité restant à traiter  */

    /* Génération d'une base de chaînes à couvrir */

    if (bytes->modifier == NULL)
    {
        bytes->raw = malloc(sizeof(sized_binary_t));
        bytes->count = 1;

        szstrdup(&bytes[0].raw[0], &bytes->orig);

        result = true;

    }
    else
        result = g_scan_token_modifier_transform(bytes->modifier, &bytes->orig, &bytes->raw, &bytes->count);

    if (!result)
        goto exit;

    /* Préparation pour la mémorisation des atomes */

    bytes->atoms = malloc(bytes->count * sizeof(tracked_scan_atom_t));

    /* Recherche des atomes */

    for (i = 0; i < bytes->count; i++)
    {
        if (bytes->flags & SPBF_CASE_INSENSITIVE)
        {
            find_best_atom(&bytes->raw[i], maxsize, &atom, &letters);

            if (letters == 0)
                bytes->atoms[i] = atom;

            /* Insertion des combinaisons pour couvrir toutes les casses */
            else
            {
                for (k = 0, extra_count = 1; k < letters; k++, extra_count *= 2)
                    ;

                extra = make_atoms_case_insensitive(&bytes->raw[i], extra_count);

                remaining = bytes->count - i - 1;

                bytes->count += (extra_count - 1);

                bytes->raw = realloc(bytes->raw, bytes->count * sizeof(sized_binary_t));

                memmove(&bytes->raw[i + extra_count], &bytes->raw[i + 1], remaining * sizeof(sized_binary_t));

                for (k = 0; k < extra_count; k++)
                    bytes->raw[i + k] = extra[k];

                free(extra);

                bytes->atoms = realloc(bytes->raw, bytes->count * sizeof(tracked_scan_atom_t));

                for (k = 0; k < extra_count; k++)
                    bytes->atoms[i + k] = atom;

                i += extra_count - 1;

            }

        }

        else
            find_best_atom(&bytes->raw[i], maxsize, &bytes->atoms[i], &letters);

    }

    /* Enregistrements en masse */


    for (i = 0; i < bytes->count && result; i++)
        result = enroll_prepared_atom(&bytes->raw[i], context, backend, &bytes->atoms[i]);

 exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : bytes   = définition de la bribe à manipuler.                *
*                context = contexte de l'analyse à mener.                     *
*                content = accès au contenu brut pour vérifications (optim.)  *
*                matches = suivi des correspondances à consolider.            *
*                                                                             *
*  Description : Transforme les correspondances locales en trouvailles.       *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_hex_bytes_check(const GScanHexBytes *bytes, GScanContext *context, GBinContent *content, pending_matches_t *matches)
{
    size_t i;                               /* Boucle de parcours #1       */
    const sized_binary_t *raw;              /* Données brutes d'origine    */
    const tracked_scan_atom_t *atom;        /* Atome correspondant         */
    size_t count;                           /* Quantité de bribes trouvées */
    const phys_t *found;                    /* Localisations des bribes    */
    size_t k;                               /* Boucle de parcours #2       */
    phys_t start;                           /* Point de départ             */
    vmpa2t pos;                             /* Position dans les données   */
    const bin_t *ptr;                       /* Accès aux données brutes    */
    int ret;                                /* Bilan d'une comparaison     */

    for (i = 0; i < bytes->count; i++)
    {
        raw = &bytes->raw[i];
        atom = &bytes->atoms[i];

        found = g_scan_context_get_atom_matches(context, atom->pid, &count);

        for (k = 0; k < count; k++)
        {
            start = found[k] - atom->pos;

            init_vmpa(&pos, start, VMPA_NO_VIRTUAL);

            /* Validation du contenu avant l'atome */

            if (atom->pos > 0)
            {
                ptr = g_binary_content_get_raw_access(content, &pos, atom->pos);

                ret = memcmp(raw->data, ptr, atom->pos);
                if (ret != 0) continue;

            }

            /* Validation du contenu après l'atome */

            if (atom->rem > 0)
            {
                advance_vmpa(&pos, atom->len);

                ptr = g_binary_content_get_raw_access(content, &pos, atom->rem);

                ret = memcmp(raw->data + atom->pos + atom->len, ptr, atom->rem);
                if (ret != 0) continue;

            }

            /* Mémorisation de la correspondance */

            add_pending_matches(matches, start, raw->len);

        }

    }

}

#endif
